%  TEST DESCRIPTION
%  Compare probability estimations obtained with {C-SVM + Platt} and our
%  proposed P-SVM solution, in 2 dimensions, on NOISY data. 
%  1) Create two normally distributed 2D datasets standing for class '1' and
%  class '-1'.
%  2) For each point, compute the probability to belong to class '1' (known from distributions).
%     ADD NOISE.
%  3) Use one dataset as labels and the other as probabilities
%     (respectively used to train C-SVM and P-SVM)
%  4) Compare results in term of probability predictions and accuracy over a
%     test set.
%     We compare:
%       1- ytest_prob : true probability to belong to class '1'
%       2- ytest_prob_svm : Platt' estimations of the probabilities
%       3- ytest_prob_svme : SVM-P estimations

%% initialisation
addpath(genpath('.'))

clear all
close all
clc

seed=1983; 
randn('seed',seed);
rand('seed',seed);

% classifier parameters
lambda = 1e-7;  
C = 100; Ct= 100;
epsilon=0.1; %0.1
kernel= 'gaussian'; %'gaussian';
kerneloption=1;
confiance = 0.025;
nu = epsilon + confiance;
a = log(1/nu -1);
plimit = nu;

%% data creation probs
nc = 100; 
ntest = 1000; 

% data set 1
sigma = 0.7; x01=[+0.3 +0.5];
% data set 2
sigma2 = 0.7; x02=[-0.3 -0.5];

% generate [X1,y1] (training data X1,training class y1 = {-1,1})
[X1,y1]=datasets_4('gaussian',nc,0,sigma, sigma2, x01, x02);
X1=single(X1);

% generate p1 (probabilities)
pc1=exp(-sum((X1-repmat(x01,nc,1)).^2,2)/(2*sigma.^2));
pc2=exp(-sum((X1-repmat(x02,nc,1)).^2,2)/(2*sigma2.^2));
yp1=pc1./(pc1+pc2);

% add noise to training data
amp = 0.1;
r = -amp + (2*amp)*rand(nc,1); 
yp1 = yp1 + r;
yp1=-(yp1<plimit)+(yp1>(1-plimit))+(yp1>plimit).*(yp1<(1-plimit)).*yp1; % proba + labeled data to avoid side effects and proba> 1 or <0

y1(yp1>0.5) = 1; y1(yp1<=0.5) = -1;

% generate [X2;y2] (test data)
[X2,y2]=datasets_4('gaussian',ntest,0,sigma, sigma2, x01, x02);
X2=single(X2);

% generate p2
clear pc1 pc2;
pc1=exp(-sum((X2-repmat(x01,ntest,1)).^2,2)/(2*sigma.^2));
pc2=exp(-sum((X2-repmat(x02,ntest,1)).^2,2)/(2*sigma2.^2));
yp2=pc1./(pc1+pc2);
ytest_prob = yp2; % only probabilities
yp2=-(yp2<plimit)+(yp2>(1-plimit))+(yp2>plimit).*(yp2<(1-plimit)).*yp2; % proba + labeled data (to avoid side effects)

% labels
y2(yp2>0.5) = 1; y2(yp2<=0.5) = -1;

% normalize training and testing  data
[Xapp, Xtest, meanxapp,stdxapp]=normalizemeanstd(X1, X2); 
yapp_svm=[y1]; % labeled data
yapp_svme=[yp1]; % proba + labeled data
ytest_lab=[y2]; % labeled data

%% Building a 2D Grid for function evaluation 
x_grid = [-1:.05:1]*(max([abs(x01(1)), abs(x02(1))]) + 2*max([sigma, sigma2]));
y_grid = [-1:0.05:1]*(max([abs(x01(2)), abs(x02(2))])+ 2*max([sigma, sigma2]));
[xtest1 xtest2]  = meshgrid(x_grid, y_grid); 
nn = length(xtest1); 
Xtest_grid = [reshape(xtest1 ,nn*nn,1) reshape(xtest2 ,nn*nn,1)]; 
nb_grid = length(Xtest_grid);
Xtest_grid_norm = (Xtest_grid - ones(nb_grid,1)*meanxapp)./ (ones(nb_grid,1)*stdxapp) ;

%% probability over the grid
clear pc1 pc2;
pc1=exp(-sum((Xtest_grid-repmat(x01,nn*nn,1)).^2,2)/(2*sigma.^2));
pc2=exp(-sum((Xtest_grid-repmat(x02,nn*nn,1)).^2,2)/(2*sigma2.^2));
ypred_grid =pc1./(pc1+pc2);
ypred_grid_2 =pc1./(pc1+pc2);
ypred_grid_2 = reshape(ypred_grid_2,nn,nn);
ypred_grid =-(ypred_grid<0.125)+(ypred_grid>0.875)+(ypred_grid>0.125).*(ypred_grid<0.875).*ypred_grid; % proba + labeled data to avoid side effects
ypred_grid = reshape(ypred_grid,nn,nn); 

%% classifier learning on a labeled problem

% training
[xsup,w,w0,pos,tps,alpha] = svmclass(Xapp,yapp_svm,C,lambda,kernel,kerneloption,1);
ypred_dist_svm = svmval(Xapp,xsup,w,w0,kernel,kerneloption,1);
ypred_lab_svm(ypred_dist_svm > 0,1) = 1; ypred_lab_svm(ypred_dist_svm <= 0,1) = -1;

% testing
ytest_dist_svm = svmval(Xtest,xsup,w,w0,kernel,kerneloption,[ones(length(Xtest),1)]);

% platt probabilities estimates
prior0 = sum(ypred_lab_svm == -1); prior1 = sum(ypred_lab_svm == 1);
[A,B]=platt(ypred_dist_svm,yapp_svm,prior0,prior1); % ytest_dist_svm => ypred_dist_svm????
ytest_prob_svm = 1./(1+exp(A*ytest_dist_svm+B));

% labels 
ytest_lab_svm(ytest_prob_svm > 0.5,1) = 1; ytest_lab_svm(ytest_prob_svm <= 0.5,1) = -1;

%--------------  Evaluating the decision function on the grid
ygrid_dist_svm = svmval(Xtest_grid_norm,xsup,w,w0,kernel,kerneloption,[ones(length(Xtest_grid),1)]);
ygrid_prob_svm = 1./(1+exp(A*ygrid_dist_svm+B));
ygrid_prob_svm = reshape(ygrid_prob_svm,nn,nn); 
ygrid_dist_svm = reshape(ygrid_dist_svm,nn,nn);

%% classifier learning with uncertainty on a semi labeled/probabilistic problem

% training
[xsup,w,w0,pos,tps,alpha] = usvmclass(Xapp,yapp_svme,C,Ct,epsilon,lambda,kernel,kerneloption,1); 
ypred_dist_svme = svmval(Xapp,xsup,w,w0,kernel,kerneloption,1);
ypred_lab_svme(ypred_dist_svm > 0,1) = 1; ypred_lab_svm(ypred_dist_svm <= 0,1) = -1;

% testing
ytest_dist_svme = svmval(Xtest,xsup,w,w0,kernel,kerneloption,[ones(length(Xtest),1)]);
ytest_prob_svme = (1./(1 + exp(-a*ytest_dist_svme)));

% labels
ytest_lab_svme(ytest_prob_svme > 0.5,1) = 1; ytest_lab_svme(ytest_prob_svme <= 0.5,1) = -1;

%--------------  Evaluating the decision function on the grid
ygrid_dist_svme = svmval(Xtest_grid_norm,xsup,w,w0,kernel,kerneloption,[ones(length(Xtest_grid),1)]);
ygrid_prob_svme = (1./(1 + exp(-a*ygrid_dist_svme)));
ygrid_prob_svme = reshape(ygrid_prob_svme,nn,nn); 
ygrid_dist_svme = reshape(ygrid_dist_svme,nn,nn); 

%% errors estimation
ixp = logical((yp2>0)&(yp2<1)); % 0.125 < proba < 0.875
ixl = logical((yp2==-1)+(yp2==1)); % proba > 0.875 ou < 0.125 => label -1/1
err_prob_svm = mean(abs(ytest_prob_svm - ytest_prob));
err_prob_svme = mean(abs(ytest_prob_svme - ytest_prob));
err_lab_svm = mean(ytest_lab_svm ~= ytest_lab);
err_lab_svme = mean(ytest_lab_svme ~= ytest_lab);
accuracy_svm = mean(ytest_lab_svm == ytest_lab);
accuracy_svme = mean(ytest_lab_svme == ytest_lab);
align_lab_svm = 1 - sum(ytest_prob_svm.*ytest_prob)/(sqrt(sum(ytest_prob_svm.^2))*sqrt(sum(ytest_prob.^2)));
align_lab_svme = 1 - sum(ytest_prob_svme.*ytest_prob)/(sqrt(sum(ytest_prob_svme.^2))*sqrt(sum(ytest_prob.^2)));
% mise_svm=sum((ytest_prob_svm-ytest_prob).^2./ytest_prob);
% mise_svme=sum((ytest_prob_svme-ytest_prob).^2./ytest_prob);
KullLeib_svm = sum(ytest_prob_svm.*log(ytest_prob_svm./ytest_prob)); % Kullback Leibler
KullLeib_svme = sum(ytest_prob_svme.*log(ytest_prob_svme./ytest_prob)); % Kullback Leibler

%% --------------- plotting platt probabilities estimates
figure(1); 
clf; 
contourf(xtest1,xtest2,ygrid_prob_svm,50);shading flat;
hold on
[cc,hh]=contour(xtest1,xtest2,ygrid_prob_svm,[0  0.5 1],'k');
clabel(cc,hh); 

h1=plot(X1(yp1>0.5,1),X1(yp1>0.5,2),'*r'); 
h2=plot(X1(yp1<=0.5,1),X1(yp1<=0.5,2),'pb'); 
title('C-SVM + Platt probability estimates');

%% --------------- plotting extended svm probabilities
figure(3); 
contourf(xtest1,xtest2,ygrid_prob_svme,50);shading flat;
hold on
[cc,hh]=contour(xtest1,xtest2,ygrid_prob_svme,[0  0.5 1],'k');
clabel(cc,hh); 

h1=plot(X1(yp1>0.5,1),X1(yp1>0.5,2),'*r'); 
h2=plot(X1(yp1<=0.5,1),X1(yp1<=0.5,2),'pb'); 
title('P-SVM probability estimates');

%% --------------- plotting true probabilities
figure(5); 
clf; 
contourf(xtest1,xtest2,ypred_grid_2,50);shading flat;
hold on
[cc,hh]=contour(xtest1,xtest2,ypred_grid_2,[0  0.5 1],'k');
clabel(cc,hh); 

h1=plot(X1(yp1>0.5,1),X1(yp1>0.5,2),'*r'); 
h2=plot(X1(yp1<=0.5,1),X1(yp1<=0.5,2),'pb'); 
title('True probabilities');



