%  TEST DESCRIPTION
%  Compare probability estimations obtained with {C-SVM + Platt} and our
%  proposed P-SVM solution, in 2 dimensions, on NOISY data with INCREASING
%  NOISE levels (random + mean)
%  1) Create two normally distributed 2D datasets standing for class '1' and
%  class '-1'.
%  2) For each point, compute the probability to belong to class '1' (known from distributions).
%     ADD NOISE.
%  3) Use one dataset as labels and the other as probabilities
%     (respectively used to train C-SVM and P-SVM)
%  4) Compare results in term of probability predictions and accuracy over a
%     test set.
%     We compare:
%       1- ytest_prob : true probability to belong to class '1'
%       2- ytest_prob_svm : Platt' estimations of the probabilities
%       3- ytest_prob_svme : SVM-P estimations
%

%% initialisation
addpath(genpath('.'))

clear all
close all
clc

seed=1985; 
randn('seed',seed);
rand('seed',seed);

% classifier parameters
lambda = 1e-7;  
C = 100; Ct= 100;
epsilon=0.1; %0.1
kernel= 'gaussian'; %'gaussian';
kerneloption=1;
confiance = 0.025;
nu = epsilon + confiance;
a = log(1/nu -1);
plimit = nu;

%% data creation probs
nc = 100; 
ntest = 1000; 

% data set 1
sigma = 0.7; x01=[+0.3 +0.5];
% data set 2
sigma2 = 0.7; x02=[-0.3 -0.5];

% generate [X1,y1] (training data X1,training class y1 = {-1,1})
[X1,y1]=datasets_4('gaussian',nc,0,sigma, sigma2, x01, x02);
X1=single(X1);

% generate p1 (probabilities)
pc1=exp(-sum((X1-repmat(x01,nc,1)).^2,2)/(2*sigma.^2));
pc2=exp(-sum((X1-repmat(x02,nc,1)).^2,2)/(2*sigma2.^2));
yp1=pc1./(pc1+pc2);

% generate [X2;y2] (test data)
[X2,y2]=datasets_4('gaussian',ntest,0,sigma, sigma2, x01, x02);
X2=single(X2);

% generate p2
clear pc1 pc2;
pc1=exp(-sum((X2-repmat(x01,ntest,1)).^2,2)/(2*sigma.^2));
pc2=exp(-sum((X2-repmat(x02,ntest,1)).^2,2)/(2*sigma2.^2));
yp2=pc1./(pc1+pc2);
ytest_prob = yp2; % only probabilities
yp2=-(yp2<plimit)+(yp2>(1-plimit))+(yp2>plimit).*(yp2<(1-plimit)).*yp2; % proba + labeled data (to avoid side effects)

% labels
y2(yp2>0.5) = 1; y2(yp2<=0.5) = -1;

% normalize training and testing  data
%[Xapp, Xtest, meanxapp,stdxapp]=normalizemeanstd(X1, X2); 
[Xapp]=normalizemeanstd(X1);
[Xtest]=normalizemeanstd(X2);
ytest_lab=[y2]; % labeled data

%% Building a 2D Grid for function evaluation
x_grid = [-1:.05:1]*(max([abs(x01(1)), abs(x02(1))]) + 2*max([sigma, sigma2]));
y_grid = [-1:0.05:1]*(max([abs(x01(2)), abs(x02(2))])+ 2*max([sigma, sigma2]));
[xtest1 xtest2]  = meshgrid(x_grid, y_grid);
nn = length(xtest1);
Xtest_grid = [reshape(xtest1 ,nn*nn,1) reshape(xtest2 ,nn*nn,1)];

%% probability over the grid
clear pc1 pc2;
pc1=exp(-sum((Xtest_grid-repmat(x01,nn*nn,1)).^2,2)/(2*sigma.^2));
pc2=exp(-sum((Xtest_grid-repmat(x02,nn*nn,1)).^2,2)/(2*sigma2.^2));
ypred_grid =pc1./(pc1+pc2);
ypred_grid_2 =pc1./(pc1+pc2);
ypred_grid_2 = reshape(ypred_grid_2,nn,nn);
ypred_grid =-(ypred_grid<plimit)+(ypred_grid>(1-plimit))+(ypred_grid>plimit).*(ypred_grid<(1-plimit)).*ypred_grid; % proba + labeled data to avoid side effects
ypred_grid = reshape(ypred_grid,nn,nn);

% add different noise to training data
range_amplitude = 0: 0.02: 0.3;
%range_amplitude = 0.1;
count_ = 0;
save_yp1 = yp1;
nb_rand = 30; 
for ampl = range_amplitude
    count_ = count_ + 1; 
    for nbr = 1:nb_rand %add nb_rand different random noises of amplitude ampl to average
        yp1 = save_yp1;
        r = -ampl + (2*ampl)*rand(nc,1);
        yp1 = yp1 + r;
        yp1=-(yp1<0.125)+(yp1>0.875)+(yp1>0.125).*(yp1<0.875).*yp1; % proba + labeled data to avoid side effects and proba> 1 or <0
        
        y1(yp1>0.5) = 1; y1(yp1<=0.5) = -1;
        
        yapp_svm=[y1]; % labeled data
        yapp_svme=[yp1]; % proba + labeled data
        
        %% classifier learning on a labeled problem
        
        % training
        [xsup,w,w0,pos,tps,alpha] = svmclass(Xapp,yapp_svm,C,lambda,kernel,kerneloption,1);
        ypred_dist_svm = svmval(Xapp,xsup,w,w0,kernel,kerneloption,1);
        ypred_lab_svm(ypred_dist_svm > 0,1) = 1; ypred_lab_svm(ypred_dist_svm <= 0,1) = -1;
        
        % testing
        ytest_dist_svm = svmval(Xtest,xsup,w,w0,kernel,kerneloption,[ones(length(Xtest),1)]);
        
        % platt probabilities estimates
        prior0 = sum(ypred_lab_svm == -1); prior1 = sum(ypred_lab_svm == 1);
        [A,B]=platt(ypred_dist_svm,yapp_svm,prior0,prior1); 
        tmp = 1./(1+exp(A*ytest_dist_svm+B));
        ytest_prob_svm(:,nbr) = tmp;
        % labels
        tmp(ytest_prob_svm(:,nbr) > 0.5,1) = 1; tmp(ytest_prob_svm(:,nbr) <= 0.5,1) = -1;
        ytest_lab_svm(:,nbr) = tmp;
        
        %--------------  Evaluating the decision function on the grid
        ygrid_dist_svm = svmval(Xtest_grid,xsup,w,w0,kernel,kerneloption,[ones(length(Xtest_grid),1)]);
        ygrid_prob_svm = 1./(1+exp(A*ygrid_dist_svm+B));
        ygrid_prob_svm = reshape(ygrid_prob_svm,nn,nn);
        ygrid_dist_svm = reshape(ygrid_dist_svm,nn,nn);
        
        %% classifier learning with uncertainty on a semi labeled/probabilistic problem
       
        % training
        [xsup,w,w0,pos,tps,alpha] = usvmclass(Xapp,yapp_svme,C,Ct,epsilon,lambda,kernel,kerneloption,1);
        ypred_dist_svme = svmval(Xapp,xsup,w,w0,kernel,kerneloption,1);
        ypred_lab_svme(ypred_dist_svm > 0,1) = 1; ypred_lab_svm(ypred_dist_svm <= 0,1) = -1;
        
        % testing
        ytest_dist_svme = svmval(Xtest,xsup,w,w0,kernel,kerneloption,[ones(length(Xtest),1)]);
        tmp = (1./(1 + exp(-a*ytest_dist_svme)));
        ytest_prob_svme(:,nbr) = tmp;
        
        % labels
        tmp(ytest_prob_svme(:,nbr) > 0.5,1) = 1; tmp(ytest_prob_svme(:,nbr) <= 0.5,1) = -1;
        ytest_lab_svme(:,nbr) = tmp;
        
        %--------------  Evaluating the decision function on the grid
        ygrid_dist_svme = svmval(Xtest_grid,xsup,w,w0,kernel,kerneloption,[ones(length(Xtest_grid),1)]);
        ygrid_prob_svme = (1./(1 + exp(-a*ygrid_dist_svme)));
        ygrid_prob_svme = reshape(ygrid_prob_svme,nn,nn);
        ygrid_dist_svme = reshape(ygrid_dist_svme,nn,nn);
    end
    %% errors estimation
    ixp = logical((yp2>0)&(yp2<1)); % 0.125 < proba < 0.875
    ixl = logical((yp2==-1)+(yp2==1)); % proba > 0.875 ou < 0.125 => label -1/1
    err_prob_svm(count_) = mean(mean(abs(ytest_prob_svm - repmat(ytest_prob,1, nb_rand))));
    err_prob_svme(count_) = mean(mean(abs(ytest_prob_svme - repmat(ytest_prob,1, nb_rand))));
    err_lab_svm(count_) = mean(mean(ytest_lab_svm ~= repmat(ytest_lab,1, nb_rand)));
    err_lab_svme(count_) = mean(mean(ytest_lab_svme ~= repmat(ytest_lab,1, nb_rand)));
    accuracy_svm(count_) = mean(mean(ytest_lab_svm == repmat(ytest_lab,1, nb_rand)));
    accuracy_svme(count_) = mean(mean(ytest_lab_svme == repmat(ytest_lab,1, nb_rand)));
    align_lab_svm(count_) = mean(1 - sum(ytest_prob_svm.*repmat(ytest_prob,1, nb_rand))./(sqrt(sum(ytest_prob_svm.^2))*sqrt(sum(ytest_prob.^2))));
    align_lab_svme(count_) = mean(1 - sum(ytest_prob_svme.*repmat(ytest_prob,1, nb_rand))/(sqrt(sum(ytest_prob_svme.^2))*sqrt(sum(ytest_prob.^2))));
    KullLeib_svm(count_) = mean(sum(ytest_prob_svm.*log(ytest_prob_svm./repmat(ytest_prob,1, nb_rand)))); % Kullback Leibler
    KullLeib_svme(count_) = mean(sum(ytest_prob_svme.*log(ytest_prob_svme./repmat(ytest_prob,1, nb_rand)))); % Kullback Leibler

end

%% display influence of noise amplitude on classification and probability estimation
figure; 
subplot(1,2,1);
hl1 = line(range_amplitude,accuracy_svm(:), 'Color', 'k'); hold on; 
hl1 = line(range_amplitude,accuracy_svme(:),'Color', 'k');
set(hl1, 'LineStyle', '--');
ax1 = gca; 
set(ax1, 'XColor', 'k', 'YColor', 'k');
ylabel('Accuracy');
xlabel('Noise amplitude');
legend('C-SVM','P-SVM');
xlim([0 .25])

subplot(1,2,2);
hl1 = line(range_amplitude,KullLeib_svm(:), 'Color', 'k'); hold on; 
hl1 = line(range_amplitude,KullLeib_svme(:),'Color', 'k');
set(hl1, 'LineStyle', '--');
ax1 = gca; 
set(ax1, 'XColor', 'k', 'YColor', 'k');
ylabel('Kullback Leibler');
xlabel('Noise amplitude');
legend('C-SVM','P-SVM');
xlim([0 .25])

